/* Copyright 2004 Garrett Rooney
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_log.h"
#include "util_filter.h"

#include "apr_xml.h"
#include "apr_date.h"

#include "genx.h"

/* $Id: mod_speedyfeed.c 401 2004-09-19 15:31:41Z rooneg $ */

typedef struct
{
  ap_filter_t *f;
  apr_bucket_brigade *bb;
} sender_baton;

static genxStatus
sender_send (void *b, constUtf8 s)
{
  apr_bucket_brigade *bb = ((sender_baton *) b)->bb;
  ap_filter_t *f = ((sender_baton *) b)->f;

  if (ap_fwrite (f, bb, s, strlen (s)) == APR_SUCCESS)
    return GENX_SUCCESS;

  return GENX_IO_ERROR;
}

static genxStatus
sender_send_bounded (void *b, constUtf8 start, constUtf8 end)
{
  apr_bucket_brigade *bb = ((sender_baton *) b)->bb;
  ap_filter_t *f = ((sender_baton *) b)->f;

  if (ap_fwrite (f, bb, start, end - start) == APR_SUCCESS)
    return GENX_SUCCESS;

  return GENX_IO_ERROR;
}

static genxStatus
sender_flush (void *b)
{
  apr_bucket_brigade *bb = ((sender_baton *) b)->bb;
  ap_filter_t *f = ((sender_baton *) b)->f;

  if (ap_filter_flush (bb, f) == APR_SUCCESS)
    return GENX_SUCCESS;

  return GENX_IO_ERROR;
}

static genxSender sender = { sender_send, sender_send_bounded, sender_flush };

static genxStatus
write_element (genxWriter w,
               const apr_xml_elem *elem,
               apr_array_header_t *namespaces,
               const char *xml_lang)
{
  apr_xml_elem *child;
  apr_xml_attr *attr;
  apr_text *t;

#define GENX_ERR(expr) do {           \
    genxStatus genx__status = (expr); \
    if (genx__status)                 \
      return genx__status;            \
  } while (0)

  GENX_ERR (genxStartElementLiteral (w,
                                     APR_XML_GET_URI_ITEM (namespaces,
                                                           elem->ns),
                                     elem->name));

  /* if we have a lang for this element and we either don't have an xml_lang
     or the xml_lang is different from this element's lang we add a new lang
     attribute and set xml_lang to the new value for anything under this. */
  if (elem->lang && (xml_lang ? strcmp (xml_lang, elem->lang) != 0 : 1))
    {
      GENX_ERR (genxAddAttributeLiteral (w,
                                         "xml",
                                         "lang",
                                         elem->lang));
      xml_lang = elem->lang;
    }

  for (attr = elem->attr; attr; attr = attr->next)
    {
      /* XXX this is an ugly hack to avoid an issue where apr's xml parsing
         code drops the xml: namespace...  man i hate that xml parser... */
      if (strncmp ("xml:", attr->name, 4) == 0)
        GENX_ERR (genxAddAttributeLiteral (w,
                                           "xml",
                                           attr->name + 4,
                                           attr->value));
      else
        GENX_ERR (genxAddAttributeLiteral (w,
                                           APR_XML_GET_URI_ITEM (namespaces,
                                                                 attr->ns),
                                           attr->name,
                                           attr->value));
    }

  for (t = elem->first_cdata.first; t; t = t->next)
    {
      GENX_ERR (genxAddText (w, t->text));
    }

  for (child = elem->first_child; child; child = child->next)
    {
      GENX_ERR (write_element (w, child, namespaces, xml_lang));

      for (t = child->following_cdata.first; t; t = t->next)
        {
          GENX_ERR (genxAddText (w, t->text));
        }
    }

  GENX_ERR (genxEndElement (w));

#undef GENX_ERR

  return GENX_SUCCESS;
}

static apr_status_t
write_document (request_rec *r,
                ap_filter_t *f,
                apr_bucket_brigade *bb,
                apr_xml_doc *pdoc)
{
  apr_status_t apr_err;
  genxStatus genx_err;
  sender_baton sb;
  genxWriter w;

  /* this is being a bit presumptuous about the encoding, but it's better
     than just leaving it off... */
  apr_err = ap_fwrite (f,
                       bb,
                       "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n",
                       39);
  if (apr_err)
    return apr_err;

  sb.f = f;
  sb.bb = bb;

  w = genxNew (NULL, NULL, &sb);
  if (! w)
    return APR_ENOMEM;

  /* declare our default namespace */
  genxDeclareNamespace (w, "http://purl.org/atom/ns#", "", &genx_err);
  if (genx_err)
    {
      ap_log_rerror (APLOG_MARK,
                     APLOG_CRIT,
                     0,
                     r,
                     "genx error %d : %s",
                     genx_err,
                     genxGetErrorMessage (w, genx_err));
 
      genxDispose (w);

      return APR_EGENERAL; 
    }

#define GENX_ERR(expr) do {                                \
    genx_err = (expr);                                     \
    if (genx_err)                                          \
      {                                                    \
        ap_log_rerror (APLOG_MARK,                         \
                       APLOG_CRIT,                         \
                       0,                                  \
                       r,                                  \
                       "genx error %d : %s",               \
                       genx_err,                           \
                       genxGetErrorMessage (w, genx_err)); \
                                                           \
        genxDispose (w);                                   \
                                                           \
        return APR_EGENERAL;                               \
      }                                                    \
  } while (0)


  GENX_ERR (genxStartDocSender (w, &sender));

  GENX_ERR (write_element (w, pdoc->root, pdoc->namespaces, NULL));

  GENX_ERR (genxEndDocument (w));

  genxDispose (w);

#undef GENX_ERR

  apr_err = ap_fwrite (f, bb, "\n", 1);
  if (apr_err)
    return apr_err;

  return APR_SUCCESS;
}

static apr_status_t
parse_time (apr_time_t *when, const char *data, apr_pool_t *pool)
{
  apr_time_exp_t exploded_time;
  const char *c = data;

  *when = APR_DATE_BAD;

  /* parse something like "2001-08-31T04:24:14Z" */
  exploded_time.tm_year = strtol (c, (char**)&c, 10);
  if (*c++ != '-')
    return APR_EINVAL;

  exploded_time.tm_mon = strtol (c, (char**)&c, 10);
  if (*c++ != '-')
    return APR_EINVAL;

  exploded_time.tm_mday = strtol (c, (char**)&c, 10);
  if (*c++ != 'T')
    return APR_EINVAL;

  exploded_time.tm_hour = strtol (c, (char**)&c, 10);
  if (*c++ != ':')
    return APR_EINVAL;

  exploded_time.tm_min = strtol (c, (char**)&c, 10);
  if (*c++ != ':')
    return APR_EINVAL;

  exploded_time.tm_sec = strtol (c, (char**)&c, 10);
  if (*c++ != 'Z')
    return APR_EINVAL;

  exploded_time.tm_usec   = 0;
  exploded_time.tm_year  -= 1900;
  exploded_time.tm_mon   -= 1;
  exploded_time.tm_wday   = 0;
  exploded_time.tm_yday   = 0;
  exploded_time.tm_isdst  = 0;
  exploded_time.tm_gmtoff = 0;

  return apr_time_exp_gmt_get (when, &exploded_time);
}

static void
drop_older_entries (apr_xml_doc *pdoc,
                    apr_time_t modified_since,
                    apr_pool_t *pool)
{
  apr_xml_elem *itr, *prev = NULL, *last_non_entry = NULL;

  for (itr = pdoc->root->first_child; itr != NULL; itr = itr->next)
    {
      if (strcmp ("entry", itr->name) == 0
          && strcmp ("http://purl.org/atom/ns#",
                     APR_XML_GET_URI_ITEM (pdoc->namespaces, itr->ns)) == 0)
        {
          apr_xml_elem *eitr;

          for (eitr = itr->first_child; eitr != NULL; eitr = eitr->next)
            {
              if (strcmp ("modified", eitr->name) == 0
                  && strcmp ("http://purl.org/atom/ns#",
                             APR_XML_GET_URI_ITEM (pdoc->namespaces,
                                                   eitr->ns)) == 0)
                {
                  apr_time_t modified;

                  parse_time (&modified,
                              eitr->first_cdata.first->text,
                              pool);

                  /* if we couldn't parse the date err on the side of wasting
                     a little bandwidth and leave it in. */
                  if (modified != APR_DATE_BAD && modified_since >= modified)
                    {
                      /* skip this one, the user already saw it. */

                      if (prev)
                        {
                          prev->next = eitr->parent->next;
                          break;
                        }
                      else
                        {
                          last_non_entry->next = eitr->parent->next;
                          break;
                        }
                    }
                  else
                    {
                      /* this is being sent down, so we want to save it for
                         later. */
                      prev = eitr->parent;
                    }
                }
            }
        }
      else
        {
          last_non_entry = itr;
        }
    }
}

typedef struct
{
  apr_xml_parser *parser;
  apr_bucket_brigade *bb;
  apr_time_t modified_since;
} speedyfeed_ctx;

static apr_status_t
speedyfeed_filter (ap_filter_t *f, apr_bucket_brigade *bb)
{
  request_rec *r = f->r;
  apr_bucket *e;

  /* we need to unset content length since we're going to change it, but if
     we do it earlier, when we're setting up the ctx, it doesn't work... */
  apr_table_unset (r->headers_out, "Content-Length");

  for (e = APR_BRIGADE_FIRST (bb);
       e != APR_BRIGADE_SENTINEL (bb);
       e = APR_BUCKET_NEXT (e))
    {
      speedyfeed_ctx *ctx = f->ctx;
      apr_status_t apr_err;
      const char *str;
      apr_size_t len;

      if (APR_BUCKET_IS_EOS (e) || APR_BUCKET_IS_FLUSH (e))
        {
          apr_xml_doc *pdoc;

          apr_err = apr_xml_parser_done (ctx->parser, &pdoc);
          if (apr_err)
            {
              ap_log_rerror (APLOG_MARK,
                             APLOG_CRIT,
                             apr_err,
                             r,
                             "couldn't finish xml parse");

              r->status_line = "500 Internal Server Error";

              return apr_err;
            }

          drop_older_entries (pdoc, ctx->modified_since, r->pool);

          apr_err = write_document (r, f->next, ctx->bb, pdoc);
          if (apr_err)
            {
              ap_log_rerror (APLOG_MARK,
                             APLOG_CRIT,
                             apr_err,
                             r,
                             "failed to write document");

              r->status_line = "500 Internal Server Error";

              return apr_err; 
            }

          APR_BUCKET_REMOVE (e);
          APR_BRIGADE_INSERT_TAIL (ctx->bb, e);
          ap_pass_brigade (f->next, ctx->bb);

          return APR_SUCCESS;
        }

      apr_err = apr_bucket_read (e, &str, &len, APR_NONBLOCK_READ);
      if (apr_err)
        {
          ap_log_rerror (APLOG_MARK,
                         APLOG_CRIT,
                         apr_err,
                         r,
                         "failed to read data from bucket");

          r->status_line = "500 Internal Server Error";

          return apr_err;
        }

      apr_err = apr_xml_parser_feed (ctx->parser, str, len);
      if (apr_err)
        {
          ap_log_rerror (APLOG_MARK,
                         APLOG_CRIT,
                         apr_err,
                         r,
                         "failed to parse feed");

          r->status_line = "500 Internal Server Error";

          return apr_err;
        }
    }

  return APR_SUCCESS;
}

static void
speedyfeed_insert_filters (request_rec *r)
{
  const char *content_type = apr_table_get (r->headers_out, "Content-Type");
  const char *if_modified_since, *a_im;

  if (! content_type)
    content_type = r->content_type;

  if (! content_type || strcmp ("application/atom+xml", content_type) != 0)
    return;

  if_modified_since = apr_table_get (r->headers_in, "If-Modified-Since");

  a_im = apr_table_get (r->headers_in, "A-IM");

  if (if_modified_since && a_im && ap_find_token (r->pool, a_im, "feed"))
    {
      speedyfeed_ctx *ctx = apr_pcalloc (r->pool, sizeof (*ctx));

      ctx->modified_since = apr_date_parse_http (if_modified_since);
      if (ctx->modified_since == APR_DATE_BAD)
        {
          ap_log_rerror (APLOG_MARK,
                         APLOG_CRIT,
                         0,
                         r,
                         "got bogus date from client '%s'",
                         if_modified_since);

          return;
        }

      ctx->bb = apr_brigade_create (r->pool, r->connection->bucket_alloc);
      ctx->parser = apr_xml_parser_create (r->pool);

      ap_add_output_filter ("SPEEDYFEED_FILTER", ctx, r, r->connection);

      /* If the status code has been added to apache use it, otherwise we
         have to fake it by setting the status line manually. */
#ifdef HTTP_IM_USED
      r->status = HTTP_IM_USED;
#else
      r->status_line = "226 IM Used";
#endif

      apr_table_set (r->headers_out, "Cache-Control", "no-store, im");
      apr_table_set (r->headers_out, "IM", "feed");

      /* Last-Modified should be set for us already */

      /* XXX should we be doing something about ETags here? */
    }
}

static void
speedyfeed_register_hooks (apr_pool_t *p)
{
  ap_hook_insert_filter (speedyfeed_insert_filters,
                         NULL,
                         NULL,
                         APR_HOOK_MIDDLE);

  ap_register_output_filter ("SPEEDYFEED_FILTER",
                             speedyfeed_filter,
                             NULL,
                             AP_FTYPE_CONTENT_SET);
}

module AP_MODULE_DECLARE_DATA speedyfeed_module =
{
  STANDARD20_MODULE_STUFF,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  speedyfeed_register_hooks
};
